@@ -1,23 +1,16 @@
 use anyhow::{bail, Error};
 use log::info;
 use odbc_api::{
-    buffers::TextRowSet, escape_attribute_value, Connection, Cursor, Environment, IntoParameter,
+    buffers::TextRowSet, escape_attribute_value, Connection, Cursor, DriverCompleteOption,
+    Environment, IntoParameter,
 };
 use std::{
-    fs::File,
+    fs::{read_to_string, File},
     io::{stdin, stdout, Read, Write},
     path::PathBuf,
 };
 use structopt::StructOpt;
 
-#[cfg(target_os = "windows")]
-use odbc_api::DriverCompleteOption;
-#[cfg(target_os = "windows")]
-use winit::{
-    event_loop::EventLoop,
-    window::{Window, WindowBuilder},
-};
-
 /// Query an ODBC data source and output the result as CSV.
 #[derive(StructOpt)]
 struct Cli {
@@ -30,11 +23,16 @@ struct Cli {
 
 #[derive(StructOpt)]
 enum Command {
-    /// Query a data source and write the result as csv.
+    /// Query a data source and write the result as csv. This is the deprecated version of `fetch`.
     Query {
         #[structopt(flatten)]
         query_opt: QueryOpt,
     },
+    /// Query a data source and write the result as csv.
+    Fetch {
+        #[structopt(flatten)]
+        fetch_opt: FetchOpt,
+    },
     /// Read the content of a csv and insert it into a table.
     Insert {
         #[structopt(flatten)]
@@ -57,7 +55,8 @@ enum Command {
     ListDataSources,
 }
 
-/// Command line arguments used to establish a connection with the ODBC data source
+// Attention: This has overwritten some help messages for the enduser if turned into a docstring:
+// Command line arguments used to establish a connection with the ODBC data source
 #[derive(StructOpt)]
 struct ConnectOpts {
     #[structopt(long, conflicts_with = "dsn")]
@@ -108,6 +107,35 @@ struct QueryOpt {
     parameters: Vec<String>,
 }
 
+#[derive(StructOpt)]
+struct FetchOpt {
+    #[structopt(flatten)]
+    connect_opts: ConnectOpts,
+    /// Number of rows queried from the database on block. Larger numbers may reduce io overhead,
+    /// but require more memory during execution.
+    #[structopt(long, default_value = "5000")]
+    batch_size: usize,
+    /// Maximum string length in bytes. If omitted no limit is applied and the ODBC driver is taken
+    /// for its word regarding the maximum length of the columns.
+    #[structopt(long, short = "m")]
+    max_str_len: Option<usize>,
+    /// Path to the output csv file the returned values are going to be written to. If omitted the
+    /// csv is going to be printed to standard out.
+    #[structopt(long, short = "o")]
+    output: Option<PathBuf>,
+    /// Query executed against the ODBC data source. Within the SQL text Question marks (`?`) can be
+    /// used as placeholders for positional parameters.
+    #[structopt(long, short = "q", conflicts_with = "sql_file")]
+    query: Option<String>,
+    /// Read the SQL query from a file, rather than a literal passed at the command line. Argument
+    /// specifies path to that file. Within the SQL text question marks (`?`) can be used as
+    /// placeholders for positional parameters.
+    #[structopt(long, short = "f", conflicts_with = "query")]
+    sql_file: Option<PathBuf>,
+    /// For each placeholder question mark (`?`) in the query text one parameter must be passed at
+    /// the end of the command line.
+    parameters: Vec<String>,
+}
 #[derive(StructOpt)]
 struct InsertOpt {
     #[structopt(flatten)]
@@ -166,7 +194,7 @@ struct ListColumnsOpt {
 
 fn main() -> Result<(), Error> {
     // Parse arguments from command line interface
-    let opt = Cli::from_args();
+    let opt = Cli::from_args_safe()?;
 
     // Initialize logging.
     stderrlog::new()
@@ -184,6 +212,9 @@ fn main() -> Result<(), Error> {
         Command::Query { query_opt } => {
             query(&environment, &query_opt)?;
         }
+        Command::Fetch { fetch_opt } => {
+            fetch(&environment, fetch_opt)?;
+        }
         Command::Insert { insert_opt } => {
             if insert_opt.batch_size == 0 {
                 bail!("batch size, must be at least 1");
@@ -233,13 +264,15 @@ fn main() -> Result<(), Error> {
 fn open_connection<'e>(
     environment: &'e Environment,
     opt: &ConnectOpts,
-) -> Result<Connection<'e>, odbc_api::Error> {
+) -> Result<Connection<'e>, Error> {
     if let Some(dsn) = opt.dsn.as_deref() {
-        return environment.connect(
-            dsn,
-            opt.user.as_deref().unwrap_or(""),
-            opt.password.as_deref().unwrap_or(""),
-        );
+        return environment
+            .connect(
+                dsn,
+                opt.user.as_deref().unwrap_or(""),
+                opt.password.as_deref().unwrap_or(""),
+            )
+            .map_err(|e| e.into());
     }
 
     // Append user and or password to connection string
@@ -252,31 +285,59 @@ fn open_connection<'e>(
     }
 
     #[cfg(target_os = "windows")]
-    if opt.prompt {
-        let window = message_only_window().unwrap();
-        return environment.driver_connect(&cs, None, DriverCompleteOption::Complete(&window));
-    }
+    let driver_completion = if opt.prompt {
+        DriverCompleteOption::Complete
+    } else {
+        DriverCompleteOption::NoPrompt
+    };
 
-    // Would rather use conditional compilation on the flag itself. While this works fine, it does
-    // mess with rust analyzer, so I keep it and panic here to keep development experience smooth.
     #[cfg(not(target_os = "windows"))]
-    if opt.prompt {
-        panic!("--prompt is only supported on windows.")
-    }
+    let driver_completion = if opt.prompt {
+        // Would rather use conditional compilation on the flag itself. While this works fine, it
+        // does mess with rust analyzer, so I keep it and panic here to keep development experience
+        // smooth.
+        bail!("--prompt is only supported on windows.");
+    } else {
+        DriverCompleteOption::NoPrompt
+    };
 
     if !opt.prompt && opt.connection_string.is_none() && opt.dsn.is_none() {
-        panic!("Either DSN, connection string or prompt must be specified.")
+        bail!("Either DSN, connection string or prompt must be specified.")
     }
 
-    environment.connect_with_connection_string(&cs)
+    environment
+        .driver_connect(&cs, None, driver_completion)
+        .map_err(|e| e.into())
 }
 
-#[cfg(target_os = "windows")]
-fn message_only_window() -> Result<Window, Error> {
-    let window = WindowBuilder::new()
-        .with_visible(false)
-        .build(&EventLoop::new())?;
-    Ok(window)
+/// Execute a query and writes the result to csv.
+fn fetch(environment: &Environment, opt: FetchOpt) -> Result<(), Error> {
+    let FetchOpt {
+        connect_opts,
+        output,
+        parameters,
+        query: query_literal,
+        batch_size,
+        max_str_len,
+        sql_file,
+    } = opt;
+
+    let query_str = match (query_literal, sql_file) {
+        (Some(literal), _) => literal,
+        (None, Some(path)) => read_to_string(path)?,
+        _ => bail!("Either `--query` or `--sql-file` must be specified."),
+    };
+
+    let query_opt = QueryOpt {
+        connect_opts,
+        batch_size,
+        max_str_len,
+        output,
+        query: query_str,
+        parameters,
+    };
+
+    query(environment, &query_opt)
 }
 
 /// Execute a query and writes the result to csv.
